@@ -1,10 +1,10 @@
 
 #==============================================================================#
 # exerb/exerb.rb
-# $Id: exerb.rb,v 1.7 2003/01/21 05:34:54 yuya Exp $
+# $Id: exerb.rb,v 1.8 2003/02/10 08:56:07 yuya Exp $
 #==============================================================================#
 
-require 'exerb/recipefile.rb'
+require 'exerb/recipefile'
 require 'exerb/archive'
 require 'exerb/archiveheader'
 require 'exerb/nameheader'
