@@ -1,4 +1,4 @@
-// $Id: cui.cpp,v 1.2 2002/12/27 06:05:31 yuya Exp $
+// $Id: cui.cpp,v 1.3 2003/01/31 03:53:19 yuya Exp $
 
 ////////////////////////////////////////////////////////////////////////////////
 
@@ -8,7 +8,7 @@
 
 int main(int argc, char** argv);
 
-static void ExOnError(LPSTR pszMessage);
+static void ExOnError(LPSTR message);
 static void ExOnFail(VALUE errinfo);
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -21,9 +21,9 @@ main(int argc, char** argv)
 
 
 static void
-ExOnError(LPSTR pszMessage)
+ExOnError(LPSTR message)
 {
-	::fprintf(stderr, "%s\n", pszMessage);
+	::fprintf(stderr, "%s\n", message);
 }
 
 static void
