@@ -1,4 +1,4 @@
-// $Id: pe.cpp,v 1.4 2003/01/31 07:22:18 yuya Exp $
+// $Id: pe.cpp,v 1.6 2003/02/20 09:18:47 yuya Exp $
 
 ////////////////////////////////////////////////////////////////////////////////
 
@@ -123,7 +123,7 @@ ExReplaceImportDllName(DWORD base_address, DWORD import_table_delta, PDWORD name
 			*name_pool_address -= length_of_dest + 1;
 			*name_pool_size    -= length_of_dest + 1;
 		} else {
-			::rb_raise(rb_eLoadError, "Fail to modify the import table. exe/dll file name is too long.");
+			::ExRaiseException(rb_eLoadError, IDS_FAIL_MODIFY_IMPORT_TABLE2);
 		}
 	}
 }
