@@ -3,10 +3,10 @@
   error.c -
 
   $Author: matz $
-  $Date: 2003/02/03 08:48:32 $
+  $Date: 2002/12/02 18:19:02 $
   created at: Mon Aug  9 16:11:34 JST 1993
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2000 Yukihiro Matsumoto
 
 **********************************************************************/
 
@@ -24,18 +24,23 @@
 #define va_init_list(a,b) va_start(a)
 #endif
 
+#if defined __CYGWIN__
+# include <cygwin/version.h>
+# if (CYGWIN_VERSION_API_MAJOR > 0) || (CYGWIN_VERSION_API_MINOR >= 8)
+#  define sys_nerr _sys_nerr
+# endif
+#endif
+
 int ruby_nerrs;
 
 static void
 err_snprintf(buf, len, fmt, args)
-    char *buf;
-    long len;
-    const char *fmt;
+    char *buf, *fmt;
+    int len;
     va_list args;
 {
-    long n;
+    int n;
 
-    ruby_set_current_source();
     if (!ruby_sourcefile) {
 	vsnprintf(buf, len, fmt, args);
 	return;
@@ -178,30 +183,30 @@ static struct types {
     int type;
     const char *name;
 } builtin_types[] = {
-    {T_NIL,	"nil"},
-    {T_OBJECT,	"Object"},
-    {T_CLASS,	"Class"},
-    {T_ICLASS,	"iClass"},	/* internal use: mixed-in module holder */
-    {T_MODULE,	"Module"},
-    {T_FLOAT,	"Float"},
-    {T_STRING,	"String"},
-    {T_REGEXP,	"Regexp"},
-    {T_ARRAY,	"Array"},
-    {T_FIXNUM,	"Fixnum"},
-    {T_HASH,	"Hash"},
-    {T_STRUCT,	"Struct"},
-    {T_BIGNUM,	"Bignum"},
-    {T_FILE,	"File"},
-    {T_TRUE,	"true"},
-    {T_FALSE,	"false"},
-    {T_SYMBOL,	"Symbol"},	/* :symbol */
-    {T_DATA,	"Data"},	/* internal use: wrapped C pointers */
-    {T_MATCH,	"MatchData"},	/* data of $~ */
-    {T_VARMAP,	"Varmap"},	/* internal use: dynamic variables */
-    {T_SCOPE,	"Scope"},	/* internal use: variable scope */
-    {T_NODE,	"Node"},	/* internal use: syntax tree node */
-    {T_UNDEF,	"undef"},	/* internal use: #undef; should not happen */
-    {-1,	0}
+    T_NIL,	"nil",
+    T_OBJECT,	"Object",
+    T_CLASS,	"Class",
+    T_ICLASS,	"iClass",	/* internal use: mixed-in module holder */
+    T_MODULE,	"Module",
+    T_FLOAT,	"Float",
+    T_STRING,	"String",
+    T_REGEXP,	"Regexp",
+    T_ARRAY,	"Array",
+    T_FIXNUM,	"Fixnum",
+    T_HASH,	"Hash",
+    T_STRUCT,	"Struct",
+    T_BIGNUM,	"Bignum",
+    T_FILE,	"File",
+    T_TRUE,	"true",
+    T_FALSE,	"false",
+    T_SYMBOL,	"Symbol",	/* :symbol */
+    T_DATA,	"Data",		/* internal use: wrapped C pointers */
+    T_MATCH,	"MatchData",	/* data of $~ */
+    T_VARMAP,	"Varmap",	/* internal use: dynamic variables */
+    T_SCOPE,	"Scope",	/* internal use: variable scope */
+    T_NODE,	"Node",		/* internal use: syntax tree node */
+    T_UNDEF,	"undef",	/* internal use: #undef; should not happen */
+    -1,		0,
 };
 
 void
@@ -230,7 +235,7 @@ rb_check_type(x, t)
 		    etype = RSTRING(rb_obj_as_string(x))->ptr;
 		}
 		else {
-		    etype = rb_obj_classname(x);
+		    etype = rb_class2name(CLASS_OF(x));
 		}
 		rb_raise(rb_eTypeError, "wrong argument type %s (expected %s)",
 			 etype, type->name);
@@ -255,13 +260,12 @@ VALUE rb_eTypeError;
 VALUE rb_eArgError;
 VALUE rb_eIndexError;
 VALUE rb_eRangeError;
-VALUE rb_eNameError;
-VALUE rb_eNoMethodError;
 VALUE rb_eSecurityError;
 VALUE rb_eNotImpError;
 VALUE rb_eNoMemError;
 
 VALUE rb_eScriptError;
+VALUE rb_eNameError;
 VALUE rb_eSyntaxError;
 VALUE rb_eLoadError;
 
@@ -274,7 +278,10 @@ rb_exc_new(etype, ptr, len)
     const char *ptr;
     long len;
 {
-    return rb_funcall(etype, rb_intern("new"), 1, rb_str_new(ptr, len));
+    VALUE exc = rb_obj_alloc(etype);
+
+    rb_iv_set(exc, "mesg", rb_str_new(ptr, len));
+    return exc;
 }
 
 VALUE
@@ -289,8 +296,11 @@ VALUE
 rb_exc_new3(etype, str)
     VALUE etype, str;
 {
-    StringValue(str);
-    return rb_exc_new(etype, RSTRING(str)->ptr, RSTRING(str)->len);
+    char *s;
+    int len;
+
+    s = rb_str2cstr(str, &len);
+    return rb_exc_new(etype, s, len);
 }
 
 static VALUE
@@ -302,10 +312,9 @@ exc_initialize(argc, argv, exc)
     VALUE mesg;
 
     if (rb_scan_args(argc, argv, "01", &mesg) == 1) {
-	StringValue(mesg);	/* ensure mesg can be converted to String */
+	STR2CSTR(mesg);		/* ensure mesg can be converted to String */
     }
     rb_iv_set(exc, "mesg", mesg);
-    rb_iv_set(exc, "bt", Qnil);
 
     return exc;
 }
@@ -316,12 +325,16 @@ exc_exception(argc, argv, self)
     VALUE *argv;
     VALUE self;
 {
-    VALUE exc;
+    VALUE etype, exc;
 
     if (argc == 0) return self;
     if (argc == 1 && self == argv[0]) return self;
-    exc = rb_obj_clone(self);
-    exc_initialize(argc, argv, exc);
+    etype = CLASS_OF(self);
+    while (FL_TEST(etype, FL_SINGLETON)) {
+	etype = RCLASS(etype)->super;
+    }
+    exc = rb_obj_alloc(etype);
+    rb_obj_call_init(exc, argc, argv);
 
     return exc;
 }
@@ -349,12 +362,12 @@ exc_inspect(exc)
 	return rb_str_dup(rb_class_path(klass));
     }
 
-    str = rb_str_buf_new2("#<");
+    str = rb_str_new2("#<");
     klass = rb_class_path(klass);
-    rb_str_buf_append(str, klass);
-    rb_str_buf_cat(str, ": ", 2);
-    rb_str_buf_append(str, exc);
-    rb_str_buf_cat(str, ">", 1);
+    rb_str_append(str, klass);
+    rb_str_cat(str, ": ", 2);
+    rb_str_append(str, exc);
+    rb_str_cat(str, ">", 1);
 
     return str;
 }
@@ -373,7 +386,7 @@ static VALUE
 check_backtrace(bt)
     VALUE bt;
 {
-    long i;
+    int i;
     static char *err = "backtrace must be Array of String";
 
     if (!NIL_P(bt)) {
@@ -400,75 +413,6 @@ exc_set_backtrace(exc, bt)
     return rb_iv_set(exc, "bt", check_backtrace(bt));
 }
 
-static VALUE
-exit_initialize(argc, argv, exc)
-    int argc;
-    VALUE *argv;
-    VALUE exc;
-{
-    VALUE status = INT2NUM(0);
-    if (argc > 0 && FIXNUM_P(argv[0])) {
-	status = *argv++;
-	--argc;
-    }
-    exc_initialize(argc, argv, exc);
-    rb_iv_set(exc, "status", status);
-    return exc;
-}
-
-static VALUE
-exit_status(exc)
-    VALUE exc;
-{
-    return rb_iv_get(exc, "status");
-}
-
-void
-#ifdef HAVE_STDARG_PROTOTYPES
-rb_name_error(ID id, const char *fmt, ...)
-#else
-rb_name_error(id, fmt, va_alist)
-    ID id;
-    const char *fmt;
-    va_dcl
-#endif
-{
-    VALUE exc;
-
-    va_list args;
-    char buf[BUFSIZ];
-
-    va_init_list(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
-    va_end(args);
-    exc = rb_exc_new2(rb_eNameError, buf);
-    rb_iv_set(exc, "name", ID2SYM(id));
-    rb_exc_raise(exc);
-}
-
-static VALUE
-name_err_name(self)
-    VALUE self;
-{
-    return rb_iv_get(self, "name");
-}
-
-static VALUE
-nometh_err_args(self)
-    VALUE self;
-{
-    return rb_iv_get(self, "args");
-}
-
-void
-rb_invalid_str(str, type)
-    const char *str, *type;
-{
-    VALUE s = rb_str_inspect(rb_str_new2(str));
-
-    rb_raise(rb_eArgError, "invalid value for %s: %s", type, RSTRING(s)->ptr);
-}
-
 static st_table *syserr_tbl;
 
 static VALUE
@@ -479,26 +423,22 @@ set_syserr(n, name)
     VALUE error;
 
     if (!st_lookup(syserr_tbl, n, &error)) {
-	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
+	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);;
 	rb_define_const(error, "Errno", INT2NUM(n));
 	st_add_direct(syserr_tbl, n, error);
     }
-    else {
-	rb_define_const(rb_mErrno, name, error);
-    }
     return error;
 }
 
 static VALUE
-get_syserr(n)
-    int n;
+get_syserr(int n)
 {
     VALUE error;
 
     if (!st_lookup(syserr_tbl, n, &error)) {
-	char name[8];	/* some Windows' errno have 5 digits. */
-
-	snprintf(name, sizeof(name), "E%03d", n);
+	char name[6];
+      
+	sprintf(name, "E%03d", n);
 	error = set_syserr(n, name);
     }
     return error;
@@ -511,28 +451,27 @@ syserr_errno(self)
     return rb_iv_get(self, "errno");
 }
 
+#ifdef __BEOS__
 static VALUE
-syserr_eqq(self, exc)
-    VALUE self, exc;
+get_syserr(int i)
 {
-    VALUE num;
-
-    if (!rb_obj_is_kind_of(exc, rb_eSystemCallError)) return Qfalse;
-    if (self == rb_eSystemCallError) return Qtrue;
-
-    num = rb_iv_get(exc, "errno");
-    if (NIL_P(num)) {
-	VALUE klass = CLASS_OF(exc);
-
-	while (TYPE(klass) == T_ICLASS || FL_TEST(klass, FL_SINGLETON)) {
-	    klass = (VALUE)RCLASS(klass)->super;
-	}
-	num = rb_const_get(klass, rb_intern("Errno"));
-    }
-    if (rb_const_get(self, rb_intern("Errno")) == num)
-	return Qtrue;
-    return Qfalse;
+   VALUE *list;
+   int ix, offset;
+   
+   if (i == B_ERROR) return syserr_error;
+   i -= B_GENERAL_ERROR_BASE;
+   ix = (i >> 12) & 0xf;
+   offset = (i >> 8) & 0xf;
+   if (offset < syserr_index[ix].n) {
+      ix = syserr_index[ix].ix;
+      if ((i & 0xff) < syserr_list[ix + offset].n) {
+	 list = syserr_list[ix + offset].list;
+	 return list[i & 0xff];
+      }
+   }
+   return 0;
 }
+#endif /* __BEOS__ */
 
 static void init_syserr _((void));
 
@@ -551,27 +490,23 @@ Init_Exception()
     rb_define_method(rb_eException, "set_backtrace", exc_set_backtrace, 1);
 
     rb_eSystemExit  = rb_define_class("SystemExit", rb_eException);
-    rb_define_method(rb_eSystemExit, "initialize", exit_initialize, -1);
-    rb_define_method(rb_eSystemExit, "status", exit_status, 0);
-
     rb_eFatal  	    = rb_define_class("fatal", rb_eException);
+    rb_eInterrupt   = rb_define_class("Interrupt", rb_eException);
     rb_eSignal      = rb_define_class("SignalException", rb_eException);
-    rb_eInterrupt   = rb_define_class("Interrupt", rb_eSignal);
 
     rb_eStandardError = rb_define_class("StandardError", rb_eException);
-    rb_eTypeError     = rb_define_class("TypeError", rb_eStandardError);
-    rb_eArgError      = rb_define_class("ArgumentError", rb_eStandardError);
-    rb_eIndexError    = rb_define_class("IndexError", rb_eStandardError);
-    rb_eRangeError    = rb_define_class("RangeError", rb_eStandardError);
-    rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
-    rb_define_method(rb_eNameError, "name", name_err_name, 0);
-    rb_eNoMethodError = rb_define_class("NoMethodError", rb_eNameError);
-    rb_define_method(rb_eNoMethodError, "args", nometh_err_args, 0);
+    rb_eTypeError   = rb_define_class("TypeError", rb_eStandardError);
+    rb_eArgError    = rb_define_class("ArgumentError", rb_eStandardError);
+    rb_eIndexError  = rb_define_class("IndexError", rb_eStandardError);
+    rb_eRangeError  = rb_define_class("RangeError", rb_eStandardError);
 
     rb_eScriptError = rb_define_class("ScriptError", rb_eException);
     rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
+    rb_eNameError   = rb_define_class("NameError", rb_eScriptError);
     rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
     rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
+    /* backward compatibility -- will be removed in the future */
+    rb_define_global_const("NotImplementError", rb_eNotImpError);
 
     rb_eRuntimeError = rb_define_class("RuntimeError", rb_eStandardError);
     rb_eSecurityError = rb_define_class("SecurityError", rb_eStandardError);
@@ -649,7 +584,7 @@ void
 rb_sys_fail(mesg)
     const char *mesg;
 {
-#if !defined(_WIN32) && !defined(__VMS)
+#ifndef NT
     char *strerror();
 #endif
     char *err;
@@ -658,10 +593,6 @@ rb_sys_fail(mesg)
     int n = errno;
     VALUE ee;
 
-    if (errno == 0) {
-	rb_bug("rb_sys_fail() - errno == 0");
-    }
-
     err = strerror(errno);
     if (mesg) {
 	volatile VALUE tmp = rb_str_inspect(rb_str_new2(mesg));
@@ -680,32 +611,6 @@ rb_sys_fail(mesg)
     rb_exc_raise(ee);
 }
 
-void
-#ifdef HAVE_STDARG_PROTOTYPES
-rb_sys_warning(const char *fmt, ...)
-#else
-rb_sys_warning(fmt, va_alist)
-     const char *fmt;
-     va_dcl
-#endif
-{
-     char buf[BUFSIZ];
-     va_list args;
-     int errno_save;
-     
-     errno_save = errno;
-
-     if (!RTEST(ruby_verbose)) return;
-
-     snprintf(buf, BUFSIZ, "warning: %s", fmt);
-     snprintf(buf+strlen(buf), BUFSIZ-strlen(buf), ": %s", strerror(errno_save));
-     
-     va_init_list(args, fmt);
-     warn_print(buf, args);
-     va_end(args);
-     errno = errno_save;
-}
-
 void
 rb_load_fail(path)
     char *path;
@@ -720,20 +625,12 @@ rb_error_frozen(what)
     rb_raise(rb_eTypeError, "can't modify frozen %s", what);
 }
 
-void
-rb_check_frozen(obj)
-    VALUE obj;
-{
-    if (OBJ_FROZEN(obj)) rb_error_frozen(rb_obj_classname(obj));
-}
-
 static void
 init_syserr()
 {
     syserr_tbl = st_init_numtable();
     rb_eSystemCallError = rb_define_class("SystemCallError", rb_eStandardError);
     rb_define_method(rb_eSystemCallError, "errno", syserr_errno, 0);
-    rb_define_singleton_method(rb_eSystemCallError, "===", syserr_eqq, 1);
 
     rb_mErrno = rb_define_module("Errno");
 #ifdef EPERM
@@ -1115,7 +1012,7 @@ err_append(s)
 	    ruby_errinfo = rb_exc_new2(rb_eSyntaxError, s);
 	}
 	else {
-	    VALUE str = rb_obj_as_string(ruby_errinfo);
+	    VALUE str = rb_str_to_str(ruby_errinfo);
 
 	    rb_str_cat2(str, "\n");
 	    rb_str_cat2(str, s);
