@@ -3,10 +3,10 @@
   class.c -
 
   $Author: matz $
-  $Date: 2002/12/19 09:32:45 $
+  $Date: 2003/01/31 04:00:14 $
   created at: Tue Aug 10 15:05:44 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
@@ -19,7 +19,7 @@
 extern st_table *rb_class_tbl;
 
 VALUE
-rb_class_new(super)
+rb_class_boot(super)
     VALUE super;
 {
     NEWOBJ(klass, struct RClass);
@@ -30,16 +30,31 @@ rb_class_new(super)
     klass->m_tbl = 0;		/* safe GC */
     klass->m_tbl = st_init_numtable();
 
+    OBJ_INFECT(klass, super);
     return (VALUE)klass;
 }
 
+VALUE
+rb_class_new(super)
+    VALUE super;
+{
+    Check_Type(super, T_CLASS);
+    if (super == rb_cClass) {
+	rb_raise(rb_eTypeError, "can't make subclass of Class");
+    }
+    if (FL_TEST(super, FL_SINGLETON)) {
+	rb_raise(rb_eTypeError, "can't make subclass of virtual class");
+    }
+    return rb_class_boot(super);
+}
+
 static int
 clone_method(mid, body, tbl)
     ID mid;
     NODE *body;
     st_table *tbl;
 {
-    st_insert(tbl, mid, NEW_METHOD(body->nd_body, body->nd_noex));
+    st_insert(tbl, mid, (st_data_t)NEW_METHOD(body->nd_body, body->nd_noex));
     return ST_CONTINUE;
 }
 
@@ -50,13 +65,20 @@ rb_mod_clone(module)
     NEWOBJ(clone, struct RClass);
     CLONESETUP(clone, module);
 
-    clone->super = RCLASS(module)->super;
+    RCLASS(clone)->super = RCLASS(module)->super;
     if (RCLASS(module)->iv_tbl) {
-	clone->iv_tbl = st_copy(RCLASS(module)->iv_tbl);
+	ID id;
+
+	RCLASS(clone)->iv_tbl = st_copy(RCLASS(module)->iv_tbl);
+	id = rb_intern("__classpath__");
+	st_delete(RCLASS(clone)->iv_tbl, &id, 0);
+	id = rb_intern("__classid__");
+	st_delete(RCLASS(clone)->iv_tbl, &id, 0);
     }
     if (RCLASS(module)->m_tbl) {
-	clone->m_tbl = st_init_numtable();
-	st_foreach(RCLASS(module)->m_tbl, clone_method, clone->m_tbl);
+	RCLASS(clone)->m_tbl = st_init_numtable();
+	st_foreach(RCLASS(module)->m_tbl, clone_method,
+	  (st_data_t)RCLASS(clone)->m_tbl);
     }
 
     return (VALUE)clone;
@@ -67,33 +89,30 @@ rb_mod_dup(mod)
     VALUE mod;
 {
     VALUE dup = rb_mod_clone(mod);
-    OBJSETUP(dup, RBASIC(mod)->klass, BUILTIN_TYPE(mod));
-    if (FL_TEST(mod, FL_SINGLETON)) {
-	FL_SET(dup, FL_SINGLETON);
-    }
+
+    RBASIC(dup)->flags = RBASIC(mod)->flags & (T_MASK|FL_TAINT|FL_SINGLETON);
     return dup;
 }
 
 VALUE
-rb_singleton_class_new(super)
-    VALUE super;
+rb_singleton_class_clone(obj)
+    VALUE obj;
 {
-    VALUE klass = rb_class_new(super);
+    VALUE klass = RBASIC(obj)->klass;
 
-    FL_SET(klass, FL_SINGLETON);
-    return klass;
-}
-
-VALUE
-rb_singleton_class_clone(klass)
-    VALUE klass;
-{
     if (!FL_TEST(klass, FL_SINGLETON))
 	return klass;
     else {
 	/* copy singleton(unnamed) class */
 	NEWOBJ(clone, struct RClass);
-	CLONESETUP(clone, klass);
+	OBJSETUP(clone, 0, RBASIC(klass)->flags);
+
+	if (BUILTIN_TYPE(obj) == T_CLASS) {
+	    RBASIC(clone)->klass = (VALUE)clone;
+	}
+	else {
+	    RBASIC(clone)->klass = rb_singleton_class_clone(klass);
+	}
 
 	clone->super = RCLASS(klass)->super;
 	clone->iv_tbl = 0;
@@ -102,7 +121,9 @@ rb_singleton_class_clone(klass)
 	    clone->iv_tbl = st_copy(RCLASS(klass)->iv_tbl);
 	}
 	clone->m_tbl = st_init_numtable();
-	st_foreach(RCLASS(klass)->m_tbl, clone_method, clone->m_tbl);
+	st_foreach(RCLASS(klass)->m_tbl, clone_method,
+	  (st_data_t)clone->m_tbl);
+	rb_singleton_class_attached(RBASIC(clone)->klass, (VALUE)clone);
 	FL_SET(clone, FL_SINGLETON);
 	return (VALUE)clone;
     }
@@ -121,12 +142,26 @@ rb_singleton_class_attached(klass, obj)
 }
 
 VALUE
-rb_make_metaclass(obj, klass)
-    VALUE obj, klass;
+rb_make_metaclass(obj, super)
+    VALUE obj, super;
 {
-    klass = rb_singleton_class_new(klass);
+    VALUE klass = rb_class_boot(super);
+    FL_SET(klass, FL_SINGLETON);
     RBASIC(obj)->klass = klass;
     rb_singleton_class_attached(klass, obj);
+    if (BUILTIN_TYPE(obj) == T_CLASS && FL_TEST(obj, FL_SINGLETON)) {
+	RBASIC(klass)->klass = klass;
+	RCLASS(klass)->super = RBASIC(rb_class_real(RCLASS(obj)->super))->klass;
+    }
+    else {
+	VALUE metasuper = RBASIC(rb_class_real(super))->klass;
+
+	/* metaclass of a superclass may be NULL at boot time */
+	if (metasuper) {
+	    RBASIC(klass)->klass = metasuper;
+	}
+    }
+
     return klass;
 }
 
@@ -171,14 +206,16 @@ rb_define_class(name, super)
 	    rb_raise(rb_eTypeError, "%s is not a class", name);
 	}
 	if (rb_class_real(RCLASS(klass)->super) != super) {
-	    rb_raise(rb_eNameError, "%s is already defined", name);
+	    rb_name_error(id, "%s is already defined", name);
 	}
 	return klass;
     }
+    if (!super) {
+	rb_warn("no super class for `%s', Object assumed", name);
+    }
     klass = rb_define_class_id(id, super);
-    rb_class_inherited(super, klass);
-
     st_add_direct(rb_class_tbl, id, klass);
+    rb_class_inherited(super, klass);
 
     return klass;
 }
@@ -199,14 +236,18 @@ rb_define_class_under(outer, name, super)
 	    rb_raise(rb_eTypeError, "%s is not a class", name);
 	}
 	if (rb_class_real(RCLASS(klass)->super) != super) {
-	    rb_raise(rb_eNameError, "%s is already defined", name);
+	    rb_name_error(id, "%s is already defined", name);
 	}
 	return klass;
     }
+    if (!super) {
+	rb_warn("no super class for `%s::%s', Object assumed",
+		rb_class2name(outer), name);
+    }
     klass = rb_define_class_id(id, super);
     rb_set_class_path(klass, outer, name);
-    rb_class_inherited(super, klass);
     rb_const_set(outer, id, klass);
+    rb_class_inherited(super, klass);
 
     return klass;
 }
@@ -252,7 +293,7 @@ rb_define_module(name)
 	module = rb_const_get(rb_cObject, id);
 	if (TYPE(module) == T_MODULE)
 	    return module;
-	rb_raise(rb_eTypeError, "%s is not a module", name);
+	rb_raise(rb_eTypeError, "%s is not a module", rb_obj_classname(module));
     }
     module = rb_define_module_id(id);
     st_add_direct(rb_class_tbl, id, module);
@@ -274,7 +315,7 @@ rb_define_module_under(outer, name)
 	if (TYPE(module) == T_MODULE)
 	    return module;
 	rb_raise(rb_eTypeError, "%s::%s is not a module",
-		 rb_class2name(outer), rb_class2name(CLASS_OF(module)));
+		 rb_class2name(outer), rb_obj_classname(module));
     }
     module = rb_define_module_id(id);
     rb_const_set(outer, id, module);
@@ -305,6 +346,8 @@ include_class_new(module, super)
     else {
 	RBASIC(klass)->klass = module;
     }
+    OBJ_INFECT(klass, module);
+    OBJ_INFECT(klass, super);
 
     return (VALUE)klass;
 }
@@ -320,6 +363,7 @@ rb_include_module(klass, module)
     if (!OBJ_TAINTED(klass)) {
 	rb_secure(4);
     }
+    
     if (NIL_P(module)) return;
     if (klass == module) return;
 
@@ -327,6 +371,7 @@ rb_include_module(klass, module)
 	Check_Type(module, T_MODULE);
     }
 
+    OBJ_INFECT(klass, module);
     c = klass;
     while (module) {
 	int superclass_seen = Qfalse;
@@ -372,6 +417,22 @@ rb_mod_included_modules(mod)
     return ary;
 }
 
+VALUE
+rb_mod_include_p(mod, mod2)
+    VALUE mod;
+    VALUE mod2;
+{
+    VALUE p;
+
+    Check_Type(mod2, T_MODULE);
+    for (p = RCLASS(mod)->super; p; p = RCLASS(p)->super) {
+	if (BUILTIN_TYPE(p) == T_ICLASS) {
+	    if (RBASIC(p)->klass == mod2) return Qtrue;
+	}
+    }
+    return Qfalse;
+}
+
 VALUE
 rb_mod_ancestors(mod)
     VALUE mod;
@@ -392,23 +453,27 @@ rb_mod_ancestors(mod)
     return ary;
 }
 
+#define VISI_CHECK(x,f) (((x)&NOEX_MASK) == (f))
+
 static int
 ins_methods_i(key, body, ary)
     ID key;
     NODE *body;
     VALUE ary;
 {
-    if ((body->nd_noex&(NOEX_PRIVATE|NOEX_PROTECTED)) == 0) {
+    if (key == ID_ALLOCATOR) return ST_CONTINUE;
+    if (!body->nd_body) {
+	rb_ary_push(ary, Qnil);
+	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+    }
+    else if (!VISI_CHECK(body->nd_noex, NOEX_PRIVATE)) {
 	VALUE name = rb_str_new2(rb_id2name(key));
 
 	if (!rb_ary_includes(ary, name)) {
-	    if (!body->nd_body) {
-		rb_ary_push(ary, Qnil);
-	    }
 	    rb_ary_push(ary, name);
 	}
     }
-    else if (body->nd_body && nd_type(body->nd_body) == NODE_ZSUPER) {
+    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
 	rb_ary_push(ary, Qnil);
 	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
     }
@@ -421,11 +486,12 @@ ins_methods_prot_i(key, body, ary)
     NODE *body;
     VALUE ary;
 {
+    if (key == ID_ALLOCATOR) return ST_CONTINUE;
     if (!body->nd_body) {
 	rb_ary_push(ary, Qnil);
 	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
     }
-    else if (body->nd_noex & NOEX_PROTECTED) {
+    else if (VISI_CHECK(body->nd_noex, NOEX_PROTECTED)) {
 	VALUE name = rb_str_new2(rb_id2name(key));
 
 	if (!rb_ary_includes(ary, name)) {
@@ -445,11 +511,37 @@ ins_methods_priv_i(key, body, ary)
     NODE *body;
     VALUE ary;
 {
+    if (key == ID_ALLOCATOR) return ST_CONTINUE;
     if (!body->nd_body) {
 	rb_ary_push(ary, Qnil);
 	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
     }
-    else if (body->nd_noex & NOEX_PRIVATE) {
+    else if (VISI_CHECK(body->nd_noex, NOEX_PRIVATE)) {
+	VALUE name = rb_str_new2(rb_id2name(key));
+
+	if (!rb_ary_includes(ary, name)) {
+	    rb_ary_push(ary, name);
+	}
+    }
+    else if (nd_type(body->nd_body) == NODE_ZSUPER) {
+	rb_ary_push(ary, Qnil);
+	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+    }
+    return ST_CONTINUE;
+}
+
+static int
+ins_methods_pub_i(key, body, ary)
+    ID key;
+    NODE *body;
+    VALUE ary;
+{
+    if (key == ID_ALLOCATOR) return ST_CONTINUE;
+    if (!body->nd_body) {
+	rb_ary_push(ary, Qnil);
+	rb_ary_push(ary, rb_str_new2(rb_id2name(key)));
+    }
+    else if (VISI_CHECK(body->nd_noex, NOEX_PUBLIC)) {
 	VALUE name = rb_str_new2(rb_id2name(key));
 
 	if (!rb_ary_includes(ary, name)) {
@@ -527,19 +619,41 @@ rb_class_private_instance_methods(argc, argv, mod)
 }
 
 VALUE
-rb_obj_singleton_methods(obj)
+rb_class_public_instance_methods(argc, argv, mod)
+    int argc;
+    VALUE *argv;
+    VALUE mod;
+{
+    VALUE option;
+
+    rb_scan_args(argc, argv, "01", &option);
+    return method_list(mod, RTEST(option), ins_methods_pub_i);
+}
+
+VALUE
+rb_obj_singleton_methods(argc, argv, obj)
+    int argc;
+    VALUE *argv;
     VALUE obj;
 {
+    VALUE all;
     VALUE ary;
     VALUE klass;
     VALUE *p, *q, *pend;
 
+    rb_scan_args(argc, argv, "01", &all);
     ary = rb_ary_new();
     klass = CLASS_OF(obj);
     while (klass && FL_TEST(klass, FL_SINGLETON)) {
 	st_foreach(RCLASS(klass)->m_tbl, ins_methods_i, ary);
 	klass = RCLASS(klass)->super;
     }
+    if (RTEST(all)) {
+	while (klass && TYPE(klass) == T_ICLASS) {
+	    st_foreach(RCLASS(klass)->m_tbl, ins_methods_i, ary);
+	    klass = RCLASS(klass)->super;
+	}
+    }
     p = q = RARRAY(ary)->ptr; pend = p + RARRAY(ary)->len;
     while (p < pend) {
 	if (*p == Qnil) {
@@ -571,10 +685,13 @@ rb_define_method(klass, name, func, argc)
     int argc;
 {
     ID id = rb_intern(name);
+    int ex = NOEX_PUBLIC;
 
-    rb_add_method(klass, id, NEW_CFUNC(func, argc), 
-		  ((name[0] == 'i' && id == rb_intern("initialize"))?
-		   NOEX_PRIVATE:NOEX_PUBLIC)|NOEX_CFUNC);
+
+    if (BUILTIN_TYPE(klass) == T_CLASS) {
+	ex |= NOEX_CFUNC;
+    }
+    rb_add_method(klass, id, NEW_CFUNC(func, argc), ex);
 }
 
 void
@@ -607,13 +724,11 @@ rb_undef_method(klass, name)
     rb_add_method(klass, rb_intern(name), 0, NOEX_UNDEF);
 }
 
-#define SPECIAL_SINGLETON(x,c) if (obj == (x)) {\
-    if (!FL_TEST(c, FL_SINGLETON)) {\
-	c = rb_singleton_class_new(c);\
-	rb_singleton_class_attached(c,obj);\
+#define SPECIAL_SINGLETON(x,c) do {\
+    if (obj == (x)) {\
+	return c;\
     }\
-    return c;\
-}
+} while (0)
 
 VALUE
 rb_singleton_class(obj)
@@ -628,17 +743,16 @@ rb_singleton_class(obj)
 	SPECIAL_SINGLETON(Qnil, rb_cNilClass);
 	SPECIAL_SINGLETON(Qfalse, rb_cFalseClass);
 	SPECIAL_SINGLETON(Qtrue, rb_cTrueClass);
-	rb_bug("unknown immediate %ld", (long)obj);
+	rb_bug("unknown immediate %ld", obj);
     }
 
     DEFER_INTS;
     if (FL_TEST(RBASIC(obj)->klass, FL_SINGLETON) &&
-       ((BUILTIN_TYPE(obj) != T_CLASS && BUILTIN_TYPE(obj) != T_MODULE) ||
-       rb_iv_get(RBASIC(obj)->klass, "__attached__") == obj)) {
+	rb_iv_get(RBASIC(obj)->klass, "__attached__") == obj) {
 	klass = RBASIC(obj)->klass;
     }
     else {
-	klass = rb_make_metaclass(obj, CLASS_OF(obj));
+	klass = rb_make_metaclass(obj, RBASIC(obj)->klass);
     }
     if (OBJ_TAINTED(obj)) {
 	OBJ_TAINT(klass);
@@ -709,11 +823,11 @@ rb_define_attr(klass, name, read, write)
 
 int
 #ifdef HAVE_STDARG_PROTOTYPES
-rb_scan_args(int argc, VALUE *argv, const char *fmt, ...)
+rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
 #else
 rb_scan_args(argc, argv, fmt, va_alist)
     int argc;
-    VALUE *argv;
+    const VALUE *argv;
     const char *fmt;
     va_dcl
 #endif
@@ -730,7 +844,7 @@ rb_scan_args(argc, argv, fmt, va_alist)
     if (ISDIGIT(*p)) {
 	n = *p - '0';
 	if (n > argc)
-	    rb_raise(rb_eArgError, "wrong # of arguments (%d for %d)", argc, n);
+	    rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)", argc, n);
 	for (i=0; i<n; i++) {
 	    var = va_arg(vargs, VALUE*);
 	    if (var) *var = argv[i];
@@ -785,7 +899,7 @@ rb_scan_args(argc, argv, fmt, va_alist)
     }
 
     if (argc > i) {
-	rb_raise(rb_eArgError, "wrong # of arguments(%d for %d)", argc, i);
+	rb_raise(rb_eArgError, "wrong number of arguments(%d for %d)", argc, i);
     }
 
     return argc;
