@@ -2,17 +2,21 @@
 
   dln.c -
 
-  $Author: knu $
-  $Date: 2002/10/17 16:27:16 $
+  $Author: michal $
+  $Date: 2003/01/16 07:34:01 $
   created at: Tue Jan 18 17:05:06 JST 1994
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
 #include "ruby.h"
 #include "dln.h"
 
+#ifdef HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
+
 #ifdef __CHECKER__
 #undef HAVE_DLOPEN
 #undef USE_DLN_A_OUT
@@ -44,7 +48,7 @@ void *xrealloc();
 #endif
 
 #include <stdio.h>
-#ifdef NT
+#if defined(_WIN32) || defined(__VMS)
 #include "missing/file.h"
 #endif
 #include <sys/types.h>
@@ -56,7 +60,8 @@ void *xrealloc();
 
 #ifdef HAVE_SYS_PARAM_H
 # include <sys/param.h>
-#else
+#endif
+#ifndef MAXPATHLEN
 # define MAXPATHLEN 1024
 #endif
 
@@ -64,10 +69,15 @@ void *xrealloc();
 # include <unistd.h>
 #endif
 
-#ifndef NT
+#ifndef _WIN32
 char *getenv();
 #endif
 
+#if defined(__VMS)
+#pragma builtins
+#include <dlfcn.h>
+#endif
+
 #ifdef __MACOS__
 # include <TextUtils.h>
 # include <CodeFragments.h>
@@ -87,36 +97,53 @@ int eaccess();
 #endif
 
 #ifndef FUNCNAME_PATTERN
-# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || defined(__OpenBSD__) || defined(NeXT) || defined(__WATCOMC__) || defined(__APPLE__)
-#  define FUNCNAME_PATTERN "_Init_%.200s"
+# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || (defined(__OpenBSD__) && !defined(__ELF__)) || defined(NeXT) || defined(__WATCOMC__) || defined(__APPLE__)
+#  define FUNCNAME_PATTERN "_Init_%s"
 # else
-#  define FUNCNAME_PATTERN "Init_%.200s"
+#  define FUNCNAME_PATTERN "Init_%s"
 # endif
 #endif
 
-static void
-init_funcname(buf, file)
-    char *buf;
-    char *file;
+static int
+init_funcname_len(buf, file)
+    char **buf;
+    const char *file;
 {
-    char *p, *slash;
+    char *p;
+    const char *slash;
+    int len;
 
     /* Load the file as an object one */
-    for (p = file, slash = p-1; *p; p++) /* Find position of last '/' */
+    for (slash = file-1; *file; file++) /* Find position of last '/' */
 #ifdef __MACOS__
-	if (*p == ':') slash = p;
+	if (*file == ':') slash = file;
 #else
-	if (*p == '/') slash = p;
+	if (*file == '/') slash = file;
 #endif
 
-    snprintf(buf, MAXPATHLEN, FUNCNAME_PATTERN, slash + 1);
-    for (p = buf; *p; p++) {         /* Delete suffix if it exists */
+    len = strlen(FUNCNAME_PATTERN) + strlen(slash + 1);
+    *buf = xmalloc(len);
+    snprintf(*buf, len, FUNCNAME_PATTERN, slash + 1);
+    for (p = *buf; *p; p++) {         /* Delete suffix if it exists */
 	if (*p == '.') {
 	    *p = '\0'; break;
 	}
     }
+    return p - *buf;
 }
 
+#define init_funcname(buf, file) do {\
+    int len = init_funcname_len(buf, file);\
+    char *tmp = ALLOCA_N(char, len+1);\
+    if (!tmp) {\
+	free(*buf);\
+	rb_memerror();\
+    }\
+    strcpy(tmp, *buf);\
+    free(*buf);\
+    *buf = tmp;\
+} while (0)
+
 #ifdef USE_DLN_A_OUT
 
 #ifndef LIBC_NAME
@@ -622,7 +649,6 @@ load_1(fd, disp, need_init)
     struct nlist *sym;
     struct nlist *end;
     int init_p = 0;
-    char buf[MAXPATHLEN];
 
     if (load_header(fd, &hdr, disp) == -1) return -1;
     if (INVALID_OBJECT(hdr)) {
@@ -631,8 +657,12 @@ load_1(fd, disp, need_init)
     }
     reloc = load_reloc(fd, &hdr, disp);
     if (reloc == NULL) return -1;
+
     syms = load_sym(fd, &hdr, disp);
-    if (syms == NULL) return -1;
+    if (syms == NULL) {
+	free(reloc);
+	return -1;
+    }
 
     sym = syms;
     end = syms + (hdr.a_syms / sizeof(struct nlist));
@@ -835,12 +865,13 @@ load_1(fd, disp, need_init)
     if (need_init) {
 	int len;
 	char **libs_to_be_linked = 0;
+	char *buf;
 
 	if (undef_tbl->num_entries > 0) {
 	    if (load_lib(libc) == -1) goto err_exit;
 	}
 
-	init_funcname(buf, need_init);
+	init_funcname(&buf, need_init);
 	len = strlen(buf);
 
 	for (sym = syms; sym<end; sym++) {
@@ -1119,6 +1150,15 @@ dln_sym(name)
 #include <windows.h>
 #endif
 
+#ifdef _WIN32_WCE
+#undef FormatMessage
+#define FormatMessage FormatMessageA
+#undef LoadLibrary
+#define LoadLibrary LoadLibraryA
+#undef GetProcAddress
+#define GetProcAddress GetProcAddressA
+#endif
+
 static const char *
 dln_strerror()
 {
@@ -1153,7 +1193,7 @@ dln_strerror()
     char *p = message;
     p += sprintf(message, "%d: ", error);
     FormatMessage(
-	FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
+	FORMAT_MESSAGE_FROM_SYSTEM	 | FORMAT_MESSAGE_IGNORE_INSERTS,
 	NULL,
 	error,
 	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
@@ -1198,7 +1238,7 @@ aix_loaderror(const char *pathname)
 #define LOAD_ERRTAB_LEN	(sizeof(load_errtab)/sizeof(load_errtab[0]))
 #define ERRBUF_APPEND(s) strncat(errbuf, s, sizeof(errbuf)-strlen(errbuf)-1)
 
-    snprintf(errbuf, 1024, "load failed - %.200s ", pathname);
+    snprintf(errbuf, 1024, "load failed - %s ", pathname);
 
     if (!loadquery(1, &message[0], sizeof(message))) 
 	ERRBUF_APPEND(strerror(errno));
@@ -1231,12 +1271,12 @@ dln_load(file)
     HINSTANCE handle;
     char winfile[MAXPATHLEN];
     void (*init_fct)();
-    char buf[MAXPATHLEN];
+    char *buf;
 
     if (strlen(file) >= MAXPATHLEN) rb_loaderror("filename too long");
 
     /* Load the file as an object one */
-    init_funcname(buf, file);
+    init_funcname(&buf, file);
 
     strcpy(winfile, file);
 
@@ -1249,6 +1289,7 @@ dln_load(file)
     if ((init_fct = (void(*)())GetProcAddress(handle, buf)) == NULL) {
 	rb_loaderror("%s - %s\n%s", dln_strerror(), buf, file);
     }
+
     /* Call the init code */
     (*init_fct)();
     return handle;
@@ -1261,9 +1302,9 @@ dln_load(file)
     return 0;
 #else
 
-    char buf[MAXPATHLEN];
+    char *buf;
     /* Load the file as an object one */
-    init_funcname(buf, file);
+    init_funcname(&buf, file);
 
 #ifdef USE_DLN_DLOPEN
 #define DLN_DEFINED
@@ -1284,13 +1325,15 @@ dln_load(file)
 	    goto failed;
 	}
 
-	if ((init_fct = (void(*)())dlsym(handle, buf)) == NULL) {
+	init_fct = (void(*)())dlsym(handle, buf);
+	if (init_fct == NULL) {
 	    error = DLN_ERROR();
 	    dlclose(handle);
 	    goto failed;
 	}
 	/* Call the init code */
 	(*init_fct)();
+
 	return handle;
     }
 #endif /* USE_DLN_DLOPEN */
@@ -1370,7 +1413,6 @@ dln_load(file)
 
 	 /* Cannot call *init_address directory, so copy this value to
 	    funtion pointer */
-
 	init_fct = (void(*)())init_address;
 	(*init_fct)();
 	return (void*)init_address;
@@ -1380,7 +1422,7 @@ dln_load(file)
 	int dyld_result;
 	NSObjectFileImage obj_file; /* handle, but not use it */
 	/* "file" is module file name .
-	   "buf" is initial function name with "_" . */
+	   "buf" is pointer to initial function name with "_" . */
 
 	void (*init_fct)();
 
@@ -1493,13 +1535,41 @@ dln_load(file)
       if (err) {
 	  rb_loaderror("Unresolved symbols - %s" , file);
       }
-	
       init_fct = (void (*)())symAddr;
       (*init_fct)();
       return (void*)init_fct;
     }
 #endif /* __MACOS__ */
 
+#if defined(__VMS)
+#define DLN_DEFINED
+    {
+	void *handle, (*init_fct)();
+	char *fname, *p1, *p2;
+
+	fname = (char *)__alloca(strlen(file)+1);
+	strcpy(fname,file);
+	if (p1 = strrchr(fname,'/'))
+	    fname = p1 + 1;
+	if (p2 = strrchr(fname,'.'))
+	    *p2 = '\0';
+
+	if ((handle = (void*)dlopen(fname, 0)) == NULL) {
+	    error = dln_strerror();
+	    goto failed;
+	}
+
+	if ((init_fct = (void (*)())dlsym(handle, buf)) == NULL) {
+	    error = DLN_ERROR();
+	    dlclose(handle);
+	    goto failed;
+	}
+	/* Call the init code */
+	(*init_fct)();
+	return handle;
+    }
+#endif /* __VMS */
+
 #ifndef DLN_DEFINED
     rb_notimplement();
 #endif
@@ -1510,6 +1580,7 @@ dln_load(file)
   failed:
     rb_loaderror("%s - %s", error, file);
 #endif
+    return 0;			/* dummy return */
 }
 
 static char *dln_find_1();
@@ -1528,7 +1599,7 @@ dln_find_exe(fname, path)
     }
 
     if (!path) {
-#if defined(MSDOS) || defined(NT) || defined(__human68k__) || defined(__MACOS__)
+#if defined(MSDOS) || defined(_WIN32) || defined(__human68k__) || defined(__MACOS__)
 	path = "/usr/local/bin;/usr/ucb;/usr/bin;/bin;.";
 #else
 	path = "/usr/local/bin:/usr/ucb:/usr/bin:/bin:.";
@@ -1590,9 +1661,8 @@ dln_find_1(fname, path, exe_flag)
     register char *dp;
     register char *ep;
     register char *bp;
-#ifndef __MACOS__
     struct stat st;
-#else
+#ifdef __MACOS__
     const char* mac_fullpath;
 #endif
 
@@ -1600,9 +1670,11 @@ dln_find_1(fname, path, exe_flag)
     if (strncmp("./", fname, 2) == 0 || strncmp("../", fname, 3) == 0)
       return fname;
     if (exe_flag && strchr(fname, '/')) return fname;
-#if defined(MSDOS) || defined(NT) || defined(__human68k__) || defined(__EMX__)
+#ifdef DOSISH
     if (fname[0] == '\\') return fname;
+# ifdef DOSISH_DRIVE_LETTER
     if (strlen(fname) > 2 && fname[1] == ':') return fname;
+# endif
     if (strncmp(".\\", fname, 2) == 0 || strncmp("..\\", fname, 3) == 0)
       return fname;
     if (exe_flag && strchr(fname, '\\')) return fname;
@@ -1632,7 +1704,7 @@ dln_find_1(fname, path, exe_flag)
 	    */
 
 	    if (*dp == '~' && (l == 1 ||
-#if defined(MSDOS) || defined(NT) || defined(__human68k__) || defined(__EMX__)
+#if defined(DOSISH)
 			       dp[1] == '\\' || 
 #endif
 			       dp[1] == '/')) {
@@ -1677,16 +1749,20 @@ dln_find_1(fname, path, exe_flag)
 	if (stat(fbuf, &st) == 0) {
 	    if (exe_flag == 0) return fbuf;
 	    /* looking for executable */
-	    if (!S_ISDIR(st.st_mode) && eaccess(fbuf, X_OK) == 0) return fbuf;
+	    if (!S_ISDIR(st.st_mode) && eaccess(fbuf, X_OK) == 0)
+		return fbuf;
 	}
 #else
 	if (mac_fullpath = _macruby_exist_file_in_libdir_as_posix_name(fbuf)) {
 	    if (exe_flag == 0) return mac_fullpath;
 	    /* looking for executable */
-	    if (eaccess(mac_fullpath, X_OK) == 0) return mac_fullpath;
+	    if (stat(mac_fullpath, &st) == 0) {
+		if (!S_ISDIR(st.st_mode) && eaccess(mac_fullpath, X_OK) == 0)
+		    return mac_fullpath;
+	    }
 	}
 #endif
-#if defined(MSDOS) || defined(NT) || defined(__human68k__) || defined(__EMX__)
+#if defined(DOSISH)
 	if (exe_flag) {
 	    static const char *extension[] = {
 #if defined(MSDOS)
@@ -1694,9 +1770,9 @@ dln_find_1(fname, path, exe_flag)
 #if defined(DJGPP)
 		".btm", ".sh", ".ksh", ".pl", ".sed",
 #endif
-#elif defined(__EMX__) || defined(NT)
+#elif defined(__EMX__) || defined(_WIN32)
 		".exe", ".com", ".cmd", ".bat",
-/* end of __EMX__ or NT*/
+/* end of __EMX__ or _WIN32 */
 #else
 		".r", ".R", ".x", ".X", ".bat", ".BAT",
 /* __human68k__ */
@@ -1722,7 +1798,7 @@ dln_find_1(fname, path, exe_flag)
 #endif
 	    }
 	}
-#endif /* MSDOS or NT or __human68k__ or __EMX__ */
+#endif /* MSDOS or _WIN32 or __human68k__ or __EMX__ */
 	/* if not, and no other alternatives, life is bleak */
 	if (*ep == '\0') {
 	    return NULL;
