@@ -2,11 +2,11 @@
 
   hash.c -
 
-  $Author: knu $
-  $Date: 2002/02/24 08:25:09 $
+  $Author: matz $
+  $Date: 2003/02/04 07:27:41 $
   created at: Mon Nov 22 18:51:18 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
   Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
   Copyright (C) 2000  Information-technology Promotion Agency, Japan
 
@@ -17,12 +17,18 @@
 #include "util.h"
 #include "rubysig.h"
 
+#ifdef __APPLE__
+#include <crt_externs.h>
+#endif
+
 #define HASH_DELETED  FL_USER1
+#define HASH_PROC_DEFAULT FL_USER2
 
 static void
 rb_hash_modify(hash)
     VALUE hash;
 {
+    if (!RHASH(hash)->tbl) rb_raise(rb_eTypeError, "uninitialized Hash");
     if (OBJ_FROZEN(hash)) rb_error_frozen("hash");
     if (!OBJ_TAINTED(hash) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify hash");
@@ -38,13 +44,13 @@ rb_hash_freeze(hash)
 VALUE rb_cHash;
 
 static VALUE envtbl;
-static ID hash;
+static ID id_hash, id_yield, id_default;
 
 VALUE
 rb_hash(obj)
     VALUE obj;
 {
-    return rb_funcall(obj, hash, 0);
+    return rb_funcall(obj, id_hash, 0);
 }
 
 static VALUE
@@ -94,9 +100,9 @@ rb_any_hash(a)
 
       default:
 	DEFER_INTS;
-	hval = rb_funcall(a, hash, 0);
+	hval = rb_funcall(a, id_hash, 0);
 	if (!FIXNUM_P(hval)) {
-	    hval = rb_funcall(hval, '%', 1, INT2FIX(65439));
+	    hval = rb_funcall(hval, '%', 1, INT2FIX(536870923));
 	}
 	ENABLE_INTS;
 	return (int)FIX2LONG(hval);
@@ -111,7 +117,7 @@ static struct st_hash_type objhash = {
 struct rb_hash_foreach_arg {
     VALUE hash;
     enum st_retval (*func)();
-    char *arg;
+    VALUE arg;
 };
 
 static int
@@ -135,7 +141,7 @@ static VALUE
 rb_hash_foreach_call(arg)
     struct rb_hash_foreach_arg *arg;
 {
-    st_foreach(RHASH(arg->hash)->tbl, rb_hash_foreach_iter, arg);
+    st_foreach(RHASH(arg->hash)->tbl, rb_hash_foreach_iter, (st_data_t)arg);
     return Qnil;
 }
 
@@ -158,7 +164,7 @@ static int
 rb_hash_foreach(hash, func, farg)
     VALUE hash;
     enum st_retval (*func)();
-    char *farg;
+    VALUE farg;
 {
     struct rb_hash_foreach_arg arg;
 
@@ -169,8 +175,9 @@ rb_hash_foreach(hash, func, farg)
     return rb_ensure(rb_hash_foreach_call, (VALUE)&arg, rb_hash_foreach_ensure, hash);
 }
 
+static VALUE hash_alloc _((VALUE));
 static VALUE
-rb_hash_new2(klass)
+hash_alloc(klass)
     VALUE klass;
 {
     NEWOBJ(hash, struct RHash);
@@ -185,19 +192,7 @@ rb_hash_new2(klass)
 VALUE
 rb_hash_new()
 {
-    return rb_hash_new2(rb_cHash);
-}
-
-static VALUE
-rb_hash_s_new(argc, argv, klass)
-    int argc;
-    VALUE *argv;
-    VALUE klass;
-{
-    VALUE hash = rb_hash_new2(klass);
-
-    rb_obj_call_init(hash, argc, argv);
-    return hash;
+    return hash_alloc(rb_cHash);
 }
 
 static VALUE
@@ -208,9 +203,18 @@ rb_hash_initialize(argc, argv, hash)
 {
     VALUE ifnone;
 
-    rb_scan_args(argc, argv, "01", &ifnone);
     rb_hash_modify(hash);
-    RHASH(hash)->ifnone = ifnone;
+    if (rb_block_given_p()) {
+	if (argc > 0) {
+	    rb_raise(rb_eArgError, "wrong number of arguments");
+	}
+	RHASH(hash)->ifnone = rb_f_lambda();
+	FL_SET(hash, HASH_PROC_DEFAULT);
+    }
+    else {
+	rb_scan_args(argc, argv, "01", &ifnone);
+	RHASH(hash)->ifnone = ifnone;
+    }
 
     return hash;
 }
@@ -225,40 +229,26 @@ rb_hash_s_create(argc, argv, klass)
     int i;
 
     if (argc == 1 && TYPE(argv[0]) == T_HASH) {
-	NEWOBJ(hash, struct RHash);
-	OBJSETUP(hash, klass, T_HASH);
+	hash = hash_alloc(klass);
 	    
-	hash->ifnone = Qnil;
-	hash->tbl = st_copy(RHASH(argv[0])->tbl);
+	RHASH(hash)->ifnone = Qnil;
+	RHASH(hash)->tbl = st_copy(RHASH(argv[0])->tbl);
 
-	return (VALUE)hash;
+	return hash;
     }
 
     if (argc % 2 != 0) {
 	rb_raise(rb_eArgError, "odd number args for Hash");
     }
-    hash = rb_hash_new2(klass);
 
+    hash = hash_alloc(klass);
     for (i=0; i<argc; i+=2) {
-	st_insert(RHASH(hash)->tbl, argv[i], argv[i+1]);
+        rb_hash_aset(hash, argv[i], argv[i + 1]);
     }
 
     return hash;
 }
 
-static VALUE
-rb_hash_clone(hash)
-    VALUE hash;
-{
-    NEWOBJ(clone, struct RHash);
-    CLONESETUP(clone, hash);
-
-    clone->ifnone = RHASH(hash)->ifnone;
-    clone->tbl = (st_table*)st_copy(RHASH(hash)->tbl);
-
-    return (VALUE)clone;
-}
-
 static VALUE
 to_hash(hash)
     VALUE hash;
@@ -281,8 +271,9 @@ rb_hash_rehash(hash)
 {
     st_table *tbl;
 
+    rb_hash_modify(hash);
     tbl = st_init_table_with_size(&objhash, RHASH(hash)->tbl->num_entries);
-    st_foreach(RHASH(hash)->tbl, rb_hash_rehash_i, tbl);
+    st_foreach(RHASH(hash)->tbl, rb_hash_rehash_i, (st_data_t)tbl);
     st_free_table(RHASH(hash)->tbl);
     RHASH(hash)->tbl = tbl;
 
@@ -296,7 +287,7 @@ rb_hash_aref(hash, key)
     VALUE val;
 
     if (!st_lookup(RHASH(hash)->tbl, key, &val)) {
-	return RHASH(hash)->ifnone;
+	return rb_funcall(hash, id_default, 1, key);
     }
     return val;
 }
@@ -315,7 +306,7 @@ rb_hash_fetch(argc, argv, hash)
     if (!st_lookup(RHASH(hash)->tbl, key, &val)) {
 	if (rb_block_given_p()) {
 	    if (argc > 1) {
-		rb_raise(rb_eArgError, "wrong # of arguments", argc);
+               rb_raise(rb_eArgError, "wrong number of arguments");
 	    }
 	    return rb_yield(key);
 	}
@@ -328,9 +319,17 @@ rb_hash_fetch(argc, argv, hash)
 }
 
 static VALUE
-rb_hash_default(hash)
+rb_hash_default(argc, argv, hash)
+    int argc;
+    VALUE *argv;
     VALUE hash;
 {
+    VALUE key;
+
+    rb_scan_args(argc, argv, "01", &key);
+    if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
+	return rb_funcall(RHASH(hash)->ifnone, id_yield, 2, hash, key);
+    }
     return RHASH(hash)->ifnone;
 }
 
@@ -340,9 +339,20 @@ rb_hash_set_default(hash, ifnone)
 {
     rb_hash_modify(hash);
     RHASH(hash)->ifnone = ifnone;
+    FL_UNSET(hash, HASH_PROC_DEFAULT);
     return ifnone;
 }
 
+static VALUE
+rb_hash_default_proc(hash)
+    VALUE hash;
+{
+    if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
+	return RHASH(hash)->ifnone;
+    }
+    return Qnil;
+}
+
 static int
 index_i(key, value, args)
     VALUE key, value;
@@ -364,7 +374,7 @@ rb_hash_index(hash, value)
     args[0] = value;
     args[1] = Qnil;
 
-    st_foreach(RHASH(hash)->tbl, index_i, args);
+    st_foreach(RHASH(hash)->tbl, index_i, (st_data_t)args);
 
     return args[1];
 }
@@ -378,15 +388,17 @@ rb_hash_indexes(argc, argv, hash)
     VALUE indexes;
     int i;
 
+    rb_warn("Hash#%s is deprecated; use Hash#select",
+	    rb_id2name(rb_frame_last_func()));
     indexes = rb_ary_new2(argc);
     for (i=0; i<argc; i++) {
 	RARRAY(indexes)->ptr[i] = rb_hash_aref(hash, argv[i]);
+	RARRAY(indexes)->len++;
     }
-    RARRAY(indexes)->len = i;
     return indexes;
 }
 
-static VALUE
+VALUE
 rb_hash_delete(hash, key)
     VALUE hash, key;
 {
@@ -404,7 +416,7 @@ rb_hash_delete(hash, key)
     if (rb_block_given_p()) {
 	return rb_yield(key);
     }
-    return RHASH(hash)->ifnone;
+    return Qnil;
 }
 
 struct shift_var {
@@ -434,13 +446,20 @@ rb_hash_shift(hash)
 
     rb_hash_modify(hash);
     var.stop = 0;
-    st_foreach(RHASH(hash)->tbl, shift_i, &var);
+    st_foreach(RHASH(hash)->tbl, shift_i, (st_data_t)&var);
 
-    if (var.stop == 0) return RHASH(hash)->ifnone;
-    return rb_assoc_new(var.key, var.val);
+    if (var.stop) {
+	return rb_assoc_new(var.key, var.val);
+    }
+    else if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
+	return rb_funcall(RHASH(hash)->ifnone, id_yield, 2, hash, Qnil);
+    }
+    else {
+	return RHASH(hash)->ifnone;
+    }
 }
 
-static int
+static enum st_retval
 delete_if_i(key, value)
     VALUE key, value;
 {
@@ -476,6 +495,42 @@ rb_hash_reject(hash)
     return rb_hash_delete_if(rb_obj_dup(hash));
 }
 
+static enum st_retval
+select_i(key, value, result)
+    VALUE key, value, result;
+{
+    VALUE assoc;
+
+    if (key == Qundef) return ST_CONTINUE;
+    assoc = rb_assoc_new(key, value);
+    if (RTEST(rb_yield(assoc)))
+	rb_ary_push(result, assoc);
+    return ST_CONTINUE;
+}
+
+VALUE
+rb_hash_select(argc, argv, hash)
+    int argc;
+    VALUE *argv;
+    VALUE hash;
+{
+    VALUE result = rb_ary_new();
+    long i;
+
+    if (rb_block_given_p()) {
+	if (argc > 0) {
+	    rb_raise(rb_eArgError, "wrong number arguments(%d for 0)", argc);
+	}
+	rb_hash_foreach(hash, select_i, result);
+    }
+    else {
+	for (i=0; i<argc; i++) {
+	    rb_ary_push(result, rb_hash_aref(hash, argv[i]));
+	}
+    }
+    return result;
+}
+
 static int
 clear_i(key, value, dummy)
     VALUE key, value, dummy;
@@ -523,8 +578,16 @@ rb_hash_replace(hash, hash2)
     VALUE hash, hash2;
 {
     hash2 = to_hash(hash2);
+    if (hash == hash2) return hash;
     rb_hash_clear(hash);
     st_foreach(RHASH(hash2)->tbl, replace_i, hash);
+    RHASH(hash)->ifnone = RHASH(hash2)->ifnone;
+    if (FL_TEST(hash2, HASH_PROC_DEFAULT)) {
+	FL_SET(hash, HASH_PROC_DEFAULT);
+    }
+    else {
+	FL_UNSET(hash, HASH_PROC_DEFAULT);
+    }
 
     return hash;
 }
@@ -545,7 +608,7 @@ rb_hash_empty_p(hash)
     return Qfalse;
 }
 
-static int
+static enum st_retval
 each_value_i(key, value)
     VALUE key, value;
 {
@@ -562,7 +625,7 @@ rb_hash_each_value(hash)
     return hash;
 }
 
-static int
+static enum st_retval
 each_key_i(key, value)
     VALUE key, value;
 {
@@ -579,7 +642,7 @@ rb_hash_each_key(hash)
     return hash;
 }
 
-static int
+static enum st_retval
 each_pair_i(key, value)
     VALUE key, value;
 {
@@ -638,11 +701,11 @@ inspect_i(key, value, str)
 	rb_str_cat2(str, ", ");
     }
     str2 = rb_inspect(key);
-    rb_str_append(str, str2);
+    rb_str_buf_append(str, str2);
     OBJ_INFECT(str, str2);
-    rb_str_cat2(str, "=>");
+    rb_str_buf_cat2(str, "=>");
     str2 = rb_inspect(value);
-    rb_str_append(str, str2);
+    rb_str_buf_append(str, str2);
     OBJ_INFECT(str, str2);
 
     return ST_CONTINUE;
@@ -654,11 +717,11 @@ inspect_hash(hash)
 {
     VALUE str;
 
-    str = rb_str_new2("{");
+    str = rb_str_buf_new2("{");
     st_foreach(RHASH(hash)->tbl, inspect_i, str);
-    rb_str_cat2(str, "}");
-    
+    rb_str_buf_cat2(str, "}");
     OBJ_INFECT(str, hash);
+
     return str;
 }
 
@@ -666,7 +729,8 @@ static VALUE
 rb_hash_inspect(hash)
     VALUE hash;
 {
-    if (RHASH(hash)->tbl->num_entries == 0) return rb_str_new2("{}");
+    if (RHASH(hash)->tbl == 0 || RHASH(hash)->tbl->num_entries == 0)
+	return rb_str_new2("{}");
     if (rb_inspecting_p(hash)) return rb_str_new2("{...}");
     return rb_protect_inspect(inspect_hash, hash, 0);
 }
@@ -767,7 +831,7 @@ rb_hash_has_value(hash, val)
 
     data[0] = Qfalse;
     data[1] = val;
-    st_foreach(RHASH(hash)->tbl, rb_hash_search_value, data);
+    st_foreach(RHASH(hash)->tbl, rb_hash_search_value, (st_data_t)data);
     return data[0];
 }
 
@@ -802,13 +866,21 @@ rb_hash_equal(hash1, hash2)
     struct equal_data data;
 
     if (hash1 == hash2) return Qtrue;
-    if (TYPE(hash2) != T_HASH) return Qfalse;
+    if (TYPE(hash2) != T_HASH) {
+	if (!rb_respond_to(hash2, rb_intern("to_hash"))) {
+	    return Qfalse;
+	}
+	return rb_equal(hash2, hash1);
+    }
     if (RHASH(hash1)->tbl->num_entries != RHASH(hash2)->tbl->num_entries)
 	return Qfalse;
+    if (!(rb_equal(RHASH(hash1)->ifnone, RHASH(hash2)->ifnone) &&
+	  FL_TEST(hash1, HASH_PROC_DEFAULT) == FL_TEST(hash2, HASH_PROC_DEFAULT)))
+	return Qfalse;
 
     data.tbl = RHASH(hash2)->tbl;
     data.result = Qtrue;
-    st_foreach(RHASH(hash1)->tbl, equal_i, &data);
+    st_foreach(RHASH(hash1)->tbl, equal_i, (st_data_t)&data);
 
     return data.result;
 }
@@ -843,24 +915,54 @@ rb_hash_update_i(key, value, hash)
     return ST_CONTINUE;
 }
 
+static int
+rb_hash_update_block_i(key, value, hash)
+    VALUE key, value;
+    VALUE hash;
+{
+    if (key == Qundef) return ST_CONTINUE;
+    if (rb_hash_has_key(hash, key)) {
+	value = rb_yield(rb_ary_new3(3, key, rb_hash_aref(hash, key), value));
+    }
+    rb_hash_aset(hash, key, value);
+    return ST_CONTINUE;
+}
+
 static VALUE
 rb_hash_update(hash1, hash2)
     VALUE hash1, hash2;
 {
     hash2 = to_hash(hash2);
-    st_foreach(RHASH(hash2)->tbl, rb_hash_update_i, hash1);
+    if (rb_block_given_p()) {
+	st_foreach(RHASH(hash2)->tbl, rb_hash_update_block_i, hash1);
+    }
+    else {
+	st_foreach(RHASH(hash2)->tbl, rb_hash_update_i, hash1);
+    }
     return hash1;
 }
 
+static VALUE
+rb_hash_merge(hash1, hash2)
+    VALUE hash1, hash2;
+{
+    return rb_hash_update(rb_obj_dup(hash1), hash2);
+}
+
 static int path_tainted = -1;
 
 static char **origenviron;
-#ifdef NT
-#define GET_ENVIRON(e) (e = win32_get_environ())
-#define FREE_ENVIRON(e) win32_free_environ(e)
+#ifdef _WIN32
+#define GET_ENVIRON(e) (e = rb_w32_get_environ())
+#define FREE_ENVIRON(e) rb_w32_free_environ(e)
 static char **my_environ;
 #undef environ
 #define environ my_environ
+#elif defined(__APPLE__)
+#undef environ
+#define environ (*_NSGetEnviron())
+#define GET_ENVIRON(e) (e)
+#define FREE_ENVIRON(e)
 #else
 extern char **environ;
 #define GET_ENVIRON(e) (e)
@@ -871,12 +973,12 @@ static VALUE
 env_delete(obj, name)
     VALUE obj, name;
 {
-    int len;
     char *nam, *val;
 
     rb_secure(4);
-    nam = rb_str2cstr(name, &len);
-    if (strlen(nam) != len) {
+    SafeStringValue(name);
+    nam = RSTRING(name)->ptr;
+    if (strlen(nam) != RSTRING(name)->len) {
 	rb_raise(rb_eArgError, "bad environment variable name");
     }
     val = getenv(nam);
@@ -884,7 +986,11 @@ env_delete(obj, name)
 	VALUE value = rb_tainted_str_new2(val);
 
 	ruby_setenv(nam, 0);
-	if (strcmp(nam, "PATH") == 0 && !OBJ_TAINTED(name)) {
+#ifdef __BORLANDC__
+	if (strcmpi(nam, "PATH") == 0) {
+#else
+	if (strcmp(nam, "PATH") == 0) {
+#endif
 	    path_tainted = 0;
 	}
 	return value;
@@ -906,15 +1012,19 @@ rb_f_getenv(obj, name)
     VALUE obj, name;
 {
     char *nam, *env;
-    int len;
 
-    nam = rb_str2cstr(name, &len);
-    if (strlen(nam) != len) {
+    StringValue(name);
+    nam = RSTRING(name)->ptr;
+    if (strlen(nam) != RSTRING(name)->len) {
 	rb_raise(rb_eArgError, "bad environment variable name");
     }
     env = getenv(nam);
     if (env) {
+#ifdef __BORLANDC__
+	if (strcmpi(nam, "PATH") == 0 && !rb_env_path_tainted())
+#else
 	if (strcmp(nam, "PATH") == 0 && !rb_env_path_tainted())
+#endif
 	    return rb_str_new2(env);
 	return rb_tainted_str_new2(env);
     }
@@ -928,18 +1038,18 @@ env_fetch(argc, argv)
 {
     VALUE key, if_none;
     char *nam, *env;
-    int len;
 
     rb_scan_args(argc, argv, "11", &key, &if_none);
-    nam = rb_str2cstr(key, &len);
-    if (strlen(nam) != len) {
+    StringValue(key);
+    nam = RSTRING(key)->ptr;
+    if (strlen(nam) != RSTRING(key)->len) {
 	rb_raise(rb_eArgError, "bad environment variable name");
     }
     env = getenv(nam);
     if (!env) {
 	if (rb_block_given_p()) {
 	    if (argc > 1) {
-		rb_raise(rb_eArgError, "wrong # of arguments", argc);
+		rb_raise(rb_eArgError, "wrong number of arguments");
 	    }
 	    return rb_yield(key);
 	}
@@ -948,7 +1058,11 @@ env_fetch(argc, argv)
 	}
 	return if_none;
     }
+#ifdef __BORLANDC__
+    if (strcmpi(nam, "PATH") == 0 && !rb_env_path_tainted())
+#else
     if (strcmp(nam, "PATH") == 0 && !rb_env_path_tainted())
+#endif
 	return rb_str_new2(env);
     return rb_tainted_str_new2(env);
 }
@@ -971,7 +1085,7 @@ rb_env_path_tainted()
 
 static int
 envix(nam)
-char *nam;
+    const char *nam;
 {
     register int i, len = strlen(nam);
     char **env;
@@ -997,44 +1111,6 @@ ruby_setenv(name, value)
     const char *value;
 {
 #if defined(WIN32) && !defined(__CYGWIN32__)
-#ifdef USE_WIN32_RTL_ENV
-    register char *envstr;
-    STRLEN namlen = strlen(name);
-    STRLEN vallen;
-    char *oldstr = environ[envix(name)];
-
-    /* putenv() has totally broken semantics in both the Borland
-     * and Microsoft CRTLs.  They either store the passed pointer in
-     * the environment without making a copy, or make a copy and don't
-     * free it. And on top of that, they dont free() old entries that
-     * are being replaced/deleted.  This means the caller must
-     * free any old entries somehow, or we end up with a memory
-     * leak every time setenv() is called.  One might think
-     * one could directly manipulate environ[], like the UNIX code
-     * above, but direct changes to environ are not allowed when
-     * calling putenv(), since the RTLs maintain an internal
-     * *copy* of environ[]. Bad, bad, *bad* stink.
-     * GSAR 97-06-07
-     */
-
-    if (!value) {
-	if (!oldstr)
-	    return;
-	value = "";
-	vallen = 0;
-    }
-    else
-	vallen = strlen(val);
-    envstr = ALLOC_N(char, namelen + vallen + 3);
-    sprintf(envstr,"%s=%s",name,value);
-    putenv(envstr);
-    if (oldstr) free(oldstr);
-#ifdef _MSC_VER
-    free(envstr);		/* MSVCRT leaks without this */
-#endif
-
-#else /* !USE_WIN32_RTL_ENV */
-
     /* The sane way to deal with the environment.
      * Has these advantages over putenv() & co.:
      *  * enables us to store a truly empty value in the
@@ -1048,10 +1124,11 @@ ruby_setenv(name, value)
      *    not see changes made by extensions that call the Win32
      *    functions directly, either.
      * GSAR 97-06-07
+     *
+     * REMARK: USE_WIN32_RTL_ENV is already obsoleted since we don't use
+     *         RTL's environ global variable directly yet.
      */
     SetEnvironmentVariable(name,value);
-#endif
-
 #elif defined __CYGWIN__
 #undef setenv
 #undef unsetenv
@@ -1120,11 +1197,10 @@ ruby_unsetenv(name)
 }
 
 static VALUE
-rb_f_setenv(obj, nm, val)
+env_aset(obj, nm, val)
     VALUE obj, nm, val;
 {
     char *name, *value;
-    int nlen, vlen;
 
     if (rb_safe_level() >= 4) {
 	rb_raise(rb_eSecurityError, "cannot change environment variable");
@@ -1135,11 +1211,13 @@ rb_f_setenv(obj, nm, val)
 	return Qnil;
     }
 
-    name = rb_str2cstr(nm, &nlen);
-    value = rb_str2cstr(val, &vlen);
-    if (strlen(name) != nlen)
+    StringValue(nm);
+    StringValue(val);
+    name = RSTRING(nm)->ptr;
+    value = RSTRING(val)->ptr;
+    if (strlen(name) != RSTRING(nm)->len)
 	rb_raise(rb_eArgError, "bad environment variable name");
-    if (strlen(value) != vlen)
+    if (strlen(value) != RSTRING(val)->len)
 	rb_raise(rb_eArgError, "bad environment variable value");
 
     ruby_setenv(name, value);
@@ -1252,7 +1330,8 @@ env_reject_bang()
 {
     volatile VALUE keys;
     VALUE *ptr;
-    int len, del = 0;
+    long len;
+    int del = 0;
 
     rb_secure(4);
     keys = env_keys();
@@ -1280,6 +1359,64 @@ env_delete_if()
     return envtbl;
 }
 
+static VALUE
+env_select(argc, argv)
+    int argc;
+    VALUE *argv;
+{
+    VALUE result = rb_ary_new();
+    long i;
+
+    if (rb_block_given_p()) {
+	char **env;
+
+	if (argc > 0) {
+	    rb_raise(rb_eArgError, "wrong number arguments(%d for 0)", argc);
+	}
+	env = GET_ENVIRON(environ);
+	while (*env) {
+	    char *s = strchr(*env, '=');
+	    if (s) {
+                VALUE assoc = rb_assoc_new(rb_tainted_str_new(*env, s-*env),
+                                           rb_tainted_str_new2(s+1));
+		if (RTEST(rb_yield(assoc))) {
+		    rb_ary_push(result, assoc);
+		}
+	    }
+	    env++;
+	}
+	FREE_ENVIRON(environ);
+    }
+    else {
+	for (i=0; i<argc; i++) {
+	    rb_ary_push(result, rb_f_getenv(Qnil, argv[i]));
+	}
+    }
+    return result;
+}
+
+static VALUE
+env_clear()
+{
+    volatile VALUE keys;
+    VALUE *ptr;
+    long len;
+    
+    rb_secure(4);
+    keys = env_keys();
+    ptr = RARRAY(keys)->ptr;
+    len = RARRAY(keys)->len; 
+
+    while (len--) {
+	VALUE val = rb_f_getenv(Qnil, *ptr);
+	if (!NIL_P(val)) {
+	    env_delete(Qnil, *ptr);
+	}
+	ptr++;
+    }
+    return envtbl;
+}
+
 static VALUE
 env_to_s()
 {
@@ -1290,7 +1427,7 @@ static VALUE
 env_inspect()
 {
     char **env;
-    VALUE str = rb_str_new2("{");
+    VALUE str = rb_str_buf_new2("{");
     VALUE i;
 
     env = GET_ENVIRON(environ);
@@ -1298,19 +1435,19 @@ env_inspect()
 	char *s = strchr(*env, '=');
 
 	if (env != environ) {
-	    rb_str_cat2(str, ", ");
+	    rb_str_buf_cat2(str, ", ");
 	}
 	if (s) {
-	    rb_str_cat2(str, "\"");
-	    rb_str_cat(str, *env, s-*env);
-	    rb_str_cat2(str, "\"=>");
+	    rb_str_buf_cat2(str, "\"");
+	    rb_str_buf_cat(str, *env, s-*env);
+	    rb_str_buf_cat2(str, "\"=>");
 	    i = rb_inspect(rb_str_new2(s+1));
-	    rb_str_append(str, i);
+	    rb_str_buf_append(str, i);
 	}
 	env++;
     }
     FREE_ENVIRON(environ);
-    rb_str_cat2(str, "}");
+    rb_str_buf_cat2(str, "}");
     OBJ_TAINT(str);
 
     return str;
@@ -1372,8 +1509,12 @@ static VALUE
 env_has_key(env, key)
     VALUE env, key;
 {
-    if (TYPE(key) != T_STRING) return Qfalse;
-    if (getenv(STR2CSTR(key))) return Qtrue;
+    char *s;
+
+    s = StringValuePtr(key);
+    if (strlen(s) != RSTRING(key)->len)
+	rb_raise(rb_eArgError, "bad environment variable name");
+    if (getenv(s)) return Qtrue;
     return Qfalse;
 }
 
@@ -1431,6 +1572,8 @@ env_indexes(argc, argv)
     int i;
     VALUE indexes = rb_ary_new2(argc);
 
+    rb_warn("ENV.%s is deprecated; use ENV.select",
+	    rb_id2name(rb_frame_last_func()));
     for (i=0;i<argc;i++) {
 	char *v = 0;
 	if (TYPE(argv[i]) == T_STRING) {
@@ -1473,20 +1616,104 @@ env_reject()
     return rb_hash_delete_if(env_to_hash());
 }
 
+static VALUE
+env_shift()
+{
+    char **env;
+
+    env = GET_ENVIRON(environ);
+    if (*env) {
+	char *s = strchr(*env, '=');
+	if (s) {
+	    VALUE key = rb_tainted_str_new(*env, s-*env);
+	    VALUE val = rb_tainted_str_new2(getenv(RSTRING(key)->ptr));
+	    env_delete(Qnil, key);
+	    return rb_assoc_new(key, val);
+	}
+    }
+    FREE_ENVIRON(environ);
+    return Qnil;
+}
+
+static VALUE
+env_invert()
+{
+    return rb_hash_invert(env_to_hash());
+}
+
+static int
+env_replace_i(key, val, keys)
+    VALUE key, val, keys;
+{
+    if (key != Qundef) {
+	env_aset(Qnil, key, val);
+	if (rb_ary_includes(keys, key)) {
+	    rb_ary_delete(keys, key);
+	}
+    }
+    return ST_CONTINUE;
+}
+
+static VALUE
+env_replace(env, hash)
+    VALUE env, hash;
+{
+    volatile VALUE keys = env_keys();
+    VALUE *ptr;
+    long len;
+
+    if (env == hash) return env;
+    hash = to_hash(hash);
+    st_foreach(RHASH(hash)->tbl, env_replace_i, keys);
+
+    ptr = RARRAY(keys)->ptr;
+    len = RARRAY(keys)->len; 
+
+    while (len--) {
+	env_delete(env, *ptr++);
+    }
+    return env;
+}
+
+static int
+env_update_i(key, val)
+    VALUE key, val;
+{
+    if (key != Qundef) {
+	if (rb_block_given_p()) {
+	    val = rb_yield(rb_ary_new3(3, key, rb_f_getenv(Qnil, key), val));
+	}
+	env_aset(Qnil, key, val);
+    }
+    return ST_CONTINUE;
+}
+
+static VALUE
+env_update(env, hash)
+    VALUE env, hash;
+{
+    if (env == hash) return env;
+    hash = to_hash(hash);
+    st_foreach(RHASH(hash)->tbl, env_update_i, 0);
+    return env;
+}
+
 void
 Init_Hash()
 {
-    hash = rb_intern("hash");
+    id_hash = rb_intern("hash");
+    id_yield = rb_intern("yield");
+    id_default = rb_intern("default");
 
     rb_cHash = rb_define_class("Hash", rb_cObject);
 
     rb_include_module(rb_cHash, rb_mEnumerable);
 
-    rb_define_singleton_method(rb_cHash, "new", rb_hash_s_new, -1);
+    rb_define_alloc_func(rb_cHash, hash_alloc);
     rb_define_singleton_method(rb_cHash, "[]", rb_hash_s_create, -1);
     rb_define_method(rb_cHash,"initialize", rb_hash_initialize, -1);
 
-    rb_define_method(rb_cHash,"clone", rb_hash_clone, 0);
+    rb_define_method(rb_cHash,"copy_object", rb_hash_replace, 1);
     rb_define_method(rb_cHash,"rehash", rb_hash_rehash, 0);
 
     rb_define_method(rb_cHash,"to_hash", rb_hash_to_hash, 0);
@@ -1499,8 +1726,9 @@ Init_Hash()
     rb_define_method(rb_cHash,"fetch", rb_hash_fetch, -1);
     rb_define_method(rb_cHash,"[]=", rb_hash_aset, 2);
     rb_define_method(rb_cHash,"store", rb_hash_aset, 2);
-    rb_define_method(rb_cHash,"default", rb_hash_default, 0);
+    rb_define_method(rb_cHash,"default", rb_hash_default, -1);
     rb_define_method(rb_cHash,"default=", rb_hash_set_default, 1);
+    rb_define_method(rb_cHash,"default_proc", rb_hash_default_proc, 0);
     rb_define_method(rb_cHash,"index", rb_hash_index, 1);
     rb_define_method(rb_cHash,"indexes", rb_hash_indexes, -1);
     rb_define_method(rb_cHash,"indices", rb_hash_indexes, -1);
@@ -1520,12 +1748,15 @@ Init_Hash()
     rb_define_method(rb_cHash,"shift", rb_hash_shift, 0);
     rb_define_method(rb_cHash,"delete", rb_hash_delete, 1);
     rb_define_method(rb_cHash,"delete_if", rb_hash_delete_if, 0);
+    rb_define_method(rb_cHash,"select", rb_hash_select, -1);
     rb_define_method(rb_cHash,"reject", rb_hash_reject, 0);
     rb_define_method(rb_cHash,"reject!", rb_hash_reject_bang, 0);
     rb_define_method(rb_cHash,"clear", rb_hash_clear, 0);
     rb_define_method(rb_cHash,"invert", rb_hash_invert, 0);
     rb_define_method(rb_cHash,"update", rb_hash_update, 1);
     rb_define_method(rb_cHash,"replace", rb_hash_replace, 1);
+    rb_define_method(rb_cHash,"merge!", rb_hash_update, 1);
+    rb_define_method(rb_cHash,"merge", rb_hash_merge, 1);
 
     rb_define_method(rb_cHash,"include?", rb_hash_has_key, 1);
     rb_define_method(rb_cHash,"member?", rb_hash_has_key, 1);
@@ -1541,17 +1772,22 @@ Init_Hash()
 
     rb_define_singleton_method(envtbl,"[]", rb_f_getenv, 1);
     rb_define_singleton_method(envtbl,"fetch", env_fetch, -1);
-    rb_define_singleton_method(envtbl,"[]=", rb_f_setenv, 2);
-    rb_define_singleton_method(envtbl,"store", rb_f_setenv, 2);
+    rb_define_singleton_method(envtbl,"[]=", env_aset, 2);
+    rb_define_singleton_method(envtbl,"store", env_aset, 2);
     rb_define_singleton_method(envtbl,"each", env_each, 0);
     rb_define_singleton_method(envtbl,"each_pair", env_each, 0);
     rb_define_singleton_method(envtbl,"each_key", env_each_key, 0);
     rb_define_singleton_method(envtbl,"each_value", env_each_value, 0);
     rb_define_singleton_method(envtbl,"delete", env_delete_m, 1);
     rb_define_singleton_method(envtbl,"delete_if", env_delete_if, 0);
+    rb_define_singleton_method(envtbl,"clear", env_clear, 0);
     rb_define_singleton_method(envtbl,"reject", env_reject, 0);
     rb_define_singleton_method(envtbl,"reject!", env_reject_bang, 0);
-    rb_define_singleton_method(envtbl,"to_s", env_to_s, 0);
+    rb_define_singleton_method(envtbl,"select", env_select, -1);
+    rb_define_singleton_method(envtbl,"shift", env_shift, 0);
+    rb_define_singleton_method(envtbl,"invert", env_invert, 0);
+    rb_define_singleton_method(envtbl,"replace", env_replace, 1);
+    rb_define_singleton_method(envtbl,"update", env_update, 1);
     rb_define_singleton_method(envtbl,"inspect", env_inspect, 0);
     rb_define_singleton_method(envtbl,"rehash", env_none, 0);
     rb_define_singleton_method(envtbl,"to_a", env_to_a, 0);
